# -*- coding: utf-8 -*-
from scipy import stats
from scipy.interpolate import interp1d
from scipy.stats.distributions import norm
from pandas import DataFrame as df
import numpy as np


def lognorm_parameters_true_to_inv(miu, sigma):
    """
    NAME: lognorm_parameters_true_to_inv
    VERSION: 0.0.1
    AUTHOR: Yan Fu, Ruben V. Coline
    DESCRIPTION:
    Converts the mean and standard deviation of distribution from x to ln(x).

    PARAMETERS:
    :param miu: True mean of the x
    :param sigma: True standard deviation of x
    :return: (miu, sigma) where miu and sigma are based on ln(x)

    USAGE:
    >>> print(lognorm_parameters_true_to_inv(0.2, 0.2))
    >>> (-1.9560115027140728, 0.8325546111576977)
    """
    cov = sigma / miu

    sigma_ln = np.sqrt(np.log(1 + cov ** 2))
    miu_ln = np.log(miu) - 1 / 2 * sigma_ln ** 2

    return miu_ln, sigma_ln


def lognorm_trunc_ppf(a, b, n_rv, sigma, loc, scale, cdf_y=None):
    """
    NAME: lognorm_trunc_ppf
    VERSION: 0.0.1
    AUTHOR: Yan Fu
    DATE: 3 Aug 2018
    DESCRIPTION:
    Truncated log normal distribution cumulative function density. Truncate and normalise  a log normal distribution
    function for a given boundary (a, b).

    PARAMETERS:
    :param a: float, Lower boundary
    :param b: float, Upper boundary
    :param n_rv: integer, total number of
    :param sigma: float, standard deviation of log normal distribution (for ln(x))
    :param loc: float, location of log normal distribution
    :param scale: float, scale of the log normal distribution
    :param cdf_y: array (1 dimension) or None. A set of numbers represent cumulative probability. If None the function
    will return the sampled values
    :return: array (1 dimension), set of numbers represent sampled values of truncated log normal distribution inline
    with 'cfd_y'

    USAGE:
    >>> import numpy as np
    >>> a = 0                   # lower boundary
    >>> b = 1                   # upper boundary
    >>> n_rv = 5                # number of random variables
    >>> sigma = 0.2             # standard deviation
    >>> miu = 0.2               # mean

    # Convert true mean and sigma to ln(x) based mean and sigma
    >>> miu, sigma = lognorm_parameters_true_to_inv(miu, sigma)

    >>> loc = np.exp(miu)       # distribution mean is 0.2 (i.e. loc = np.exp(miu))
    >>> scale = 1               # default
    >>> result = lognorm_trunc_ppf(a, b, n_rv, sigma, loc, scale)
    >>> print(result)
    [0.14142136 0.49653783 0.65783987 0.81969479 1.        ]
    """

    # Generate a linear spaced array inline with lower and upper boundary of log normal cumulative probability density.
    sampled_cfd = np.linspace(
        stats.lognorm.cdf(x=a, s=sigma, loc=loc, scale=scale),
        stats.lognorm.cdf(x=b, s=sigma, loc=loc, scale=scale),
        n_rv
    )

    # Sample log normal distribution
    sampled = stats.lognorm.ppf(q=sampled_cfd, s=sigma, loc=loc, scale=scale)

    return sampled


def gumbel_parameter_converter(miu, sigma):
    """
    NAME: gumbel_parameter_converter
    VERSION: 0.0.1
    AUTHOR: Yan Fu, Ruben V. Coline
    DATE: 27 Sept 2018
    DESCRIPTION:
    This function is used in conjunction of the scipy.stats.gumbel distribution, converts mean and standard deviation
    (sigma) of samples x to location and scale which is used in scipy.stats.gumbel_r function.

    :param miu: mean value of samples x
    :param sigma: standard deviation of samples x
    :return: location and scale

    EXAMPLE:
    """

    # parameters Gumbel W&S
    alpha = 1.282 / sigma
    u = miu - 0.5772 / alpha

    # parameters Gumbel scipy
    scale = 1 / alpha
    loc = u

    return loc, scale


def gumbel_parameter_converter2(loc, scale):
    # parameters Gumbel W&S
    # alpha = 1.282 / sigma
    # u = miu - 0.5772 / alpha

    sigma = 1.282 / (1 / scale)
    miu = loc + 0.5772 / (1 / scale)

    # parameters Gumbel scipy
    # scale = 1 / alpha
    # loc = u

    return miu, sigma


def gumbel_r_trunc_ppf(a, b, n_rv, loc, scale, cdf_y=None):
    """
    NAME: gumbel_r_trunc_ppf
    VERSION: 0.0.1
    AUTHOR: Yan Fu
    DATE: 27 Sept 2018
    DESCRIPTION: Produces evenly sampled random variables based on gumbel distribution (tail to the x+ direction, i.e.
    median greater than mean). Truncation is possible via variables 'a' and 'b'. i.e. inversed cumulative density
    function f(x), x will be sampled in linear space ranging from 'a' to 'b'. Then f(x) is returned. Additionally, if x
    is defined 'cdf_y' then f(cdf_y) is returned.

    PARAMETERS:
    :param a: lower bound of truncation
    :param b: upper bound of truncation
    :param n_rv: number of random variables to be sampled, equal to the length of the returned array
    :param loc: location of the distribution
    :param scale: scale of the distribution
    :param cdf_y: array ranging with range (0, 1)
    :return sampled: sampled random variables

    USAGE:
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> loc, scale = gumbel_parameter_converter(600, 180)
    >>> rv = gumbel_r_trunc_ppf(0, 1800, 10, loc, scale)
    >>> print(np.round(rv, 0))
    [   0.  408.  462.  506.  548.  594.  646.  713.  819. 1800.]

    """

    # Generate a linear spaced array inline with lower and upper boundary of log normal cumulative probability density.
    sampled_cfd = np.linspace(
        stats.gumbel_r.cdf(x=a, loc=loc, scale=scale),
        stats.gumbel_r.cdf(x=b, loc=loc, scale=scale),
        n_rv
    )

    # Following three lines are used to check the validity of the distribution
    # print("511 (0.80): {:.4f}".format(stats.gumbel_r._y_(x=510, loc=loc, scale=scale)))
    # print("584 (0.90): {:.4f}".format(stats.gumbel_r._y_(x=584, loc=loc, scale=scale)))
    # print("655 (0.95): {:.4f}".format(stats.gumbel_r._y_(x=655, loc=loc, scale=scale)))

    # Sample log normal distribution
    sampled = stats.gumbel_r.ppf(q=sampled_cfd, loc=loc, scale=scale)

    # Work out cumulative probability function from 'sampled', output in forms of x y.
    # Interpolate x and y are processed to be capable to cope with two extreme values. y[0] (cumulative probability,
    # initial boundary) is manually set to 0.
    x = np.linspace(a, b, int(n_rv) + 1, endpoint=False)
    x += (x[1] - x[0]) / 2
    x = x[0:-2]
    y = np.array([np.sum(sampled <= i) for i in x]) / len(sampled)

    # Interpolate
    f = interp1d(y, x, bounds_error=False, fill_value=(np.min(y), np.max(y)))

    if cdf_y is None:
        return sampled
    else:
        return f(cdf_y)


def latin_hypercube_sampling(num_samples, num_arguments=1, sample_lbound=0, sample_ubound=1):
    """
    NAME: latin_hypercube_sampling
    AUTHOR: Yan Fu
    VERSION: 0.1
    DATE: 3 Aug 2018
    DESCRIPTION:
    Latin Hypercube Sampling, generates an nxm array where m equal to 'num_arguments' and n equal to 'num_samples'.
    Current version only adopts 'centered' sampling mode (each sampled cell value is centered).

    PARAMETERS:
    :param num_samples: Number of samples (i.e. rows)
    :param num_arguments: Number of arguments (i.e. columns)
    :param sample_lbound: Lower sampling boundary
    :param sample_ubound: Upper sampling boundary
    :return: An array with shape (num_samples, num_arguments)

    EXAMPLE:
    >>> result = latin_hypercube_sampling(num_samples=10, num_arguments=3, sample_lbound=0, sample_ubound=0.001)
    This example yields an array with shape of (100, 3), with each column filled 100 linear spaced numbers (shuffled)
    from 1 to 0.001 (i.e. shuffled [0.05, 0.15, 0.25, 0.35, 0.45, 0.55, 0.65, 0.75, 0.85, 0.95]).
    An example output:
    result = [
        [0.85 0.95 0.55]
        [0.55 0.45 0.05]
        [0.25 0.25 0.15]
        [0.35 0.85 0.75]
        [0.45 0.65 0.85]
        [0.75 0.05 0.65]
        [0.05 0.35 0.45]
        [0.15 0.15 0.35]
        [0.65 0.55 0.25]
        [0.95 0.75 0.95]
    ]
    """

    if sample_lbound > sample_ubound:
        sample_ubound += sample_lbound
        sample_lbound = sample_ubound - sample_lbound
        sample_ubound = sample_ubound - sample_lbound

    # Generate sorted integers with correct shape
    mat_random_num = np.linspace(sample_lbound, sample_ubound, num_samples + 1, dtype=float)
    mat_random_num += (mat_random_num[1] - mat_random_num[0]) * 0.5
    mat_random_num = mat_random_num[0:-1]
    mat_random_num = np.reshape(mat_random_num, (len(mat_random_num), 1))
    mat_random_nums = mat_random_num * np.ones((1, num_arguments))

    # np.random.shuffle(mat_random_nums)

    for i in range(np.shape(mat_random_nums)[1]):
        np.random.shuffle(mat_random_nums[:, i])

    if num_arguments == 1:
        mat_random_nums = mat_random_nums.flatten()

    return mat_random_nums


def mc_inputs_generator_worker(arg):
    kwargs, q = arg
    result = mc_inputs_generator(**kwargs)
    q.put("index")
    return result


def mc_inputs_generator(
        n_simulations,
        room_depth,
        room_opening_fraction_lbound,
        room_opening_fraction_ubound,
        room_opening_fraction_std,
        room_opening_fraction_mean,
        room_opening_permanent_fraction,
        fire_qfd_std,
        fire_qfd_mean,
        fire_qfd_ubound,
        fire_qfd_lbound,
        fire_hrr_density_std,
        fire_hrr_density_mean,
        fire_hrr_density_ubound,
        fire_hrr_density_lbound,
        fire_com_eff_lbound,
        fire_com_eff_ubound,
        fire_spread_lbound,
        fire_spread_ubound,
        fire_nft_mean,
        beam_loc_ratio_lbound,
        beam_loc_ratio_ubound,
        **_
):

    # ==================================================================================================================
    # CHECKS
    # ==================================================================================================================

    # Fire duration has to be as long as travelling fire to travel through the entire floor
    # time_end = np.max([time_end, room_depth / fire_spread_lbound])

    # ==================================================================================================================
    # Distribution variables
    # ==================================================================================================================
    df_input_samples = None

    if n_simulations == 1:

        df_input_samples = df(
            dict(
                window_open_fraction=room_opening_fraction_mean,
                fire_load_density=fire_qfd_mean,
                fire_spread_speed=(fire_spread_ubound + fire_spread_lbound) / 2,
                beam_position=room_depth * 0.75,
                fire_nft_ubound=fire_nft_mean,
                fire_hrr_density=fire_hrr_density_mean,
                index=0
            )
        )
        df_input_samples.set_index("index", inplace=True)

    elif n_simulations == 2:

        raise ValueError('Number of simulations need to be greater than 2.')

    elif n_simulations > 2:

        def generate_samples(dist_type:str, dist_scale:float, dist_loc:float, lbound:float, ubound:float, n:int):

            cfd_q = np.linspace(
                getattr(stats, dist_type).cdf(x=lbound, loc=dist_loc, scale=dist_scale),
                getattr(stats, dist_type).cdf(x=ubound, loc=dist_loc, scale=dist_scale),
                n
            )
            samples = getattr(stats, dist_type).ppf(q=cfd_q, loc=dist_loc, scale=dist_scale)
            samples[samples == np.inf] = ubound
            samples[samples == -np.inf] = lbound
            np.random.shuffle(samples)

            return samples

        # Fuel load density
        # -----------------
        fire_combustion_efficiency = np.linspace(fire_com_eff_lbound, fire_com_eff_ubound, n_simulations)
        qfd_loc, qfd_scale = gumbel_parameter_converter(fire_qfd_mean, fire_qfd_std)
        fire_load_density_samples = generate_samples('gumbel_r', qfd_scale, qfd_loc, fire_qfd_ubound, fire_qfd_lbound, n_simulations) * fire_combustion_efficiency

        # Fire HRR density (travelling fire)
        # --------------------------------------------------------------------------------------------------------------
        fire_hrr_density_samples = generate_samples('norm', fire_hrr_density_std, fire_hrr_density_mean, fire_hrr_density_ubound, fire_hrr_density_lbound, n_simulations)

        # Opening fraction factor (glazing fall-out fraction)
        # --------------------------------------------------------------------------------------------------------------
        opening_fraction_mean_conv, opening_fraction_std_conv = lognorm_parameters_true_to_inv(
            room_opening_fraction_mean,
            room_opening_fraction_std
        )
        opening_sigma = opening_fraction_std_conv
        opening_loc = 0
        openg_scale = np.exp(opening_fraction_mean_conv)
        window_open_fraction_samples = 1 - lognorm_trunc_ppf(room_opening_fraction_lbound, room_opening_fraction_ubound, n_simulations, opening_sigma, opening_loc, openg_scale)
        window_open_fraction_samples[window_open_fraction_samples == np.inf] = room_opening_fraction_ubound
        window_open_fraction_samples[window_open_fraction_samples == -np.inf] = room_opening_fraction_lbound
        window_open_fraction_samples = window_open_fraction_samples * (1 - room_opening_permanent_fraction) + room_opening_permanent_fraction
        np.random.shuffle(window_open_fraction_samples)

        # Beam location
        # --------------------------------------------------------------------------------------------------------------
        beam_position_samples = np.linspace(beam_loc_ratio_lbound, beam_loc_ratio_ubound, n_simulations) * room_depth
        np.random.shuffle(beam_position_samples)

        # Fire spread speed (travelling fire)
        # --------------------------------------------------------------------------------------------------------------
        fire_spread_speed_samples = np.linspace(fire_spread_lbound, fire_spread_ubound, n_simulations)
        np.random.shuffle(fire_spread_speed_samples)

        # Near field temperature (travelling fire)
        # --------------------------------------------------------------------------------------------------------------
        # todo: check lower and upper limit values
        fire_nft_std = (1.939 - (np.log(fire_nft_mean) * 0.266)) * fire_nft_mean
        fire_nft_ubound_samples = generate_samples('norm', fire_nft_std, fire_nft_mean, 500, 1500, n_simulations)

        # Summary
        # --------------------------------------------------------------------------------------------------------------
        df_input_samples = df(
            dict(
                fire_load_density=fire_load_density_samples,
                fire_spread_speed=fire_spread_speed_samples,
                fire_nft_ubound=fire_nft_ubound_samples,
                fire_hrr_density=fire_hrr_density_samples,
                fire_combustion_effeciency=fire_combustion_efficiency,
                beam_position=beam_position_samples,
                window_open_fraction=window_open_fraction_samples,
                index=np.arange(0, n_simulations, 1)
            )
        )
        df_input_samples.set_index("index", inplace=True)

    return df_input_samples


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    pass



# -*- coding: utf-8 -*-

import warnings
import numpy as np
import copy
from sfeprapy.func.heat_transfer_1d.solver_ndarray import solver_solid, solver_exposed


def c_steel_T(temperature):

    temperature -= 273.15

    if temperature < 20:
        warnings.warn('Temperature ({:.1f} °C) is below 20 °C'.format(temperature))
        return 425 + 0.773 * 20 - 1.69e-3 * np.power(20, 2) + 2.22e-6 * np.power(20, 3)
    if 20 <= temperature < 600:
        return 425 + 0.773 * temperature - 1.69e-3 * np.power(temperature, 2) + 2.22e-6 * np.power(temperature, 3)
    elif 600 <= temperature < 735:
        return 666 + 13002 / (738 - temperature)
    elif 735 <= temperature < 900:
        return 545 + 17820 / (temperature - 731)
    elif 900 <= temperature <= 1200:
        return 650
    elif temperature > 1200:
        warnings.warn('Temperature ({:.1f} °C) is greater than 1200 °C'.format(temperature))
        return 650
    else:
        warnings.warn('Temperature ({:.1f} °C) is outside bound.'.format(temperature))
        return 0

from typing import Union


def steel_c_T_vectorised(T: np.ndarray):

    T = T - 273.15  # unit conversion: K -> deg.C

    c = np.where(T < 20, 425, 0)
    c = np.where((20 <= T) & (T <= 600), 425 + 0.773 * T - 1.69e-3 * np.power(T, 2) + 2.22e-6 * np.power(T, 3), c)
    c = np.where((600 < T) & (T <= 735), 666 + 13002 / (738 - T), c)
    c = np.where((735 < T) & (T <= 900), 545 + 17820 / (T - 731), c)
    c = np.where((900 < T) & (T <= 1200), 650, c)
    c = np.where(1200 < T, 651, c)

    return c


def k_steel_T(temperature):
    temperature -= 273.15
    if temperature < 20:
        warnings.warn('Temperature ({:.1f} °C) is below 20 °C'.format(temperature))
        return 54 - 3.33e-2 * 20
    if temperature < 800:
        return 54 - 3.33e-2 * temperature
    elif temperature <= 1200:
        return 27.3
    elif temperature > 1200:
        warnings.warn('Temperature ({:.1f} °C) is greater than 1200 °C'.format(temperature))
        return 27.3
    else:
        warnings.warn('Temperature ({:.1f} °C) is outside bound.'.format(temperature))
        return 0


def steel_k_T_vectorised(T: np.ndarray):

    T = T - 273.15  # unit conversion: K -> deg.C

    k = np.where(T < 20, 54, 0)
    k = np.where((20 <= T) & (T <= 800), 54 - 3.33e-2 * T, k)
    k = np.where((800 < T) & (T <= 1200), 27.3, k)
    k = np.where(1200 < T, 27.3, k)

    return k


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import seaborn as sns

    # Solid properties
    thickness = 0.1
    m = 20
    rho = 7850  # kg/m3
    c_func = c_steel_T  # J/kg/K
    k_func = k_steel_T  # W/K/m

    # Boundary properties
    U0 = 1000 + 273.15
    h0 = 25
    U1 = 20 + 273.15
    h1 = 4
    emissivity = 0.7

    # Time properties
    t0 = 0
    t1 = 600

    # construct vectors
    thickness += (thickness / m * 2)

    m += 2

    x = np.linspace(0, thickness, 2*m+1)[1::2]  # location of nodes
    dx = np.full((m,), thickness/m)  # thickness of nodes
    U = np.ones_like(dx) * (20 + 273.15)  # temperature of nodes
    rho = np.full_like(dx, rho)
    h = np.zeros_like(dx)  # convection coefficients


    e = np.zeros_like(dx)  # radiation heat transfer coefficient
    e[[0, -1]] = 5.67e-8 * emissivity  # radiation at gas nodes

    x = np.array([])
    x -= (x[0] * 2)
    dx = np.array([])
    dx[[0, -1]] = 1  # gas nodes have unity thickness
    rho
    h
    e
    h[0], h[-1] = h0, h1  # convection at gas nodes

    # ==================================================================================================================
    # ONE-DIMENSIONAL FINITE DIFFERENCE START
    # ==================================================================================================================

    c = np.full_like(dx, 460)
    # k = np.full_like(dx, 30)

    dU_dt_ = list()  # results container - temperature change rate
    U_ = list()
    t_ = list()

    t = t0  # set simulation time
    while t <= t1:
        print(int(t))

        if int(t) == 300:
            print('d')

        c = steel_c_T_vectorised(U)
        # k = np.fromiter(map(k_steel_T, U), dtype=float) / dx
        k = steel_k_T_vectorised(U) / dx
        k[[0, -1]] = 0

        dt = 0.005  # constant time step

        # Temperatures at boundary (i.e. gas)
        # ===================================

        i0_exposed = 1
        i1_exposed = -1

        U[:i0_exposed], U[i1_exposed:] = U0, U1  # boundary at constant temperature

        # Temperatures at (solid) outer nodes, only energy in and out, no into other solids
        # =================================================================================

        U[1] = solver_exposed(
            dx=dx[1],
            U_solid=U[1],
            U_gas=U[0],
            rho=rho[1],
            c=c[1],
            h=h[0],
            e=e[0],
            dt=dt,
        )

        # u0 = U[0] - U[1]
        # u1 = 0 # U[1] - U[2]
        # u0_4 = U[0]**4 - U[1]**4
        # u1_4 = U[1]**4 - U[2]**4
        # e0 = u0 * h[0] - u1 * h[1]  # convection
        # e1 = u0 * k[0] - u1 * k[1]  # conduction
        # e2 = u0_4 * e[0] - u1_4 * e[1]  # radiation
        # e3 = rho[1] * c[1] * dx[1]  # stored
        # dU_dt_0 = (e0 + e1 + e2) / e3
        # U[1] = U[1] + dU_dt_0 * dt

        # Temperatures at (solid) outer nodes, only energy in and out, no into other solids
        # =================================================================================

        U[-2] = solver_exposed(
            dx=dx[-2],
            U_solid=U[-2],
            U_gas=U[-1],
            rho=rho[-2],
            c=c[-2],
            h=h[-1],
            e=e[-1],
            dt=dt,
        )

        # u0 = 0 # U[-3] - U[-2]
        # u1 = U[-2] - U[-1]
        # u0_4 = U[-3]**4 - U[-2]**4
        # u1_4 = U[-2]**4 - U[-1]**4
        # e0 = u0 * h[-2] - u1 * h[-1]  # convection
        # e1 = u0 * k[-2] - u1 * k[-1]  # conduction
        # e2 = u0_4 * e[-2] - u1_4 * e[-1]  # radiation
        # e3 = rho[-2] * c[-2] * dx[-2]  # stored
        # dU_dt_1 = (e0 + e1 + e2) / e3
        # U[-2] = U[-2] + dU_dt_1 * dt

        # Temperatures at intermediate nodes
        # ==================================

        U[1:-1] = solver_solid(
            dx=dx[1:-1],
            U=U[1:-1],
            k=k[1:-1],
            rho=rho[1:-1],
            c=c[1:-1],
            dt=dt,
        )

        # for m in range(2, len(dx)-2, 1):
        #     du0 = U[m-1] - U[m]
        #     du1 = U[m] - U[m+1]
        #     e1 = du0 * k[m] - du1 * k[m]
        #     en = rho[m] * c[m] * dx[m]
        #     du_dt = e1 / en
        #     U[m] += du_dt * dt

        # u0 = U[1:-3] - U[2:-2]
        # u1 = U[2:-2] - U[3:-1]
        # e0 = u0 * h[1:-3] - u1 * h[3:-1]
        # e1 = u0 * k[2:-2] - u1 * k[2:-2]
        # e3 = rho[2:-2] * c[2:-2] * dx[2:-2] ** 2
        # dU_dt = (e0 + e1) / e3
        # U[2:-2] = U[2:-2] + dU_dt * dt

        # Post stuff
        # ==========

        U_.append(copy.copy(U))
        t_.append(t)

        # Increment time
        # ==============

        t += dt

    fig, ax1 = plt.subplots()

    for i in range(0, len(U_), int(len(U_)/10)):
        u = U_[i]
        sns.lineplot(x[1:-1], u[1:-1] - 273.15, ax=ax1, label=round(t_[i], 7))

    if t_[i] != t_[-1]:
        sns.lineplot(x[1:-1], U[1:-1]-273.15, ax=ax1, label=int(t))

    plt.tight_layout()
    plt.show()





def fire(time, growth_factor, cap_hrr=0, cap_hrr_to_time=0):
    if isinstance(growth_factor, str):
        growth_factor = str.lower(growth_factor)
        growth_dict = {"slow": 0.0029,
                       "medium": 0.0117,
                       "fast": 0.0469,
                       "ultra-fast": 0.1876}
        try:
            growth_factor = growth_dict[growth_factor]
        except KeyError:
            err_msg = "{} should be one of the following if not a number: {}"
            err_msg = err_msg.format("growth_factor", ", ".join(list(growth_dict.keys())))
            raise ValueError(err_msg)

    heat_release_rate = growth_factor * time ** 2 * 1000

    # cap hrr
    if cap_hrr:
        heat_release_rate[heat_release_rate>cap_hrr] = cap_hrr

    if cap_hrr_to_time:
        heat_release_rate[time>cap_hrr_to_time] = -1
        heat_release_rate[heat_release_rate==-1] = max(heat_release_rate)

    return heat_release_rate



# -*- coding: utf-8 -*-
import json
import multiprocessing as mp
import os
import time
import warnings
from tkinter import filedialog, Tk, StringVar

import numpy as np
import pandas as pd
from tqdm import tqdm

from sfeprapy.func.fire_iso834 import fire as _fire_standard
from sfeprapy.mc0.mc0_func_gen import mc_inputs_generator
from sfeprapy.mc0.mc0_func_main import calc_time_equivalence_worker, calc_time_equivalence, y_results_summary


def select_path_input_csv():
    """get a list of dict()s representing different scenarios"""

    root = Tk()
    root.withdraw()
    folder_path = StringVar()

    path_input_file_csv = filedialog.askopenfile(
        title='Select Input File',
        parent=root,
        filetypes=[('SPREADSHEET', ['.csv', '.xlsx'])],
        mode='r'
    )

    folder_path.set(path_input_file_csv)

    root.update()

    try:
        path_input_file_csv = os.path.realpath(path_input_file_csv.name)
        return path_input_file_csv
    except AttributeError:
        raise FileNotFoundError('file not found.')


def mc_inputs_dict_to_df(dict_in: dict):
    """DESCRIPTION
    Generate sampled values for stochastic parameters based upon their distributions, this process includes static
    parameters where they are duplicated for a pre-defined number (i.e. number of simulations).

    PARAMETERS
    :param dict_in: dict, with keys defining parameter name and value defining their value.
    :return df_mc: DataFrame, sampled set as rows and parameters as columns.
    """

    df_mc = mc_inputs_generator(index=True, **dict_in)

    n_simulations = len(df_mc.index)
    beam_c = 0
    iso834_time = np.arange(0, 6 * 60 * 60, 30)
    iso834_temperature = _fire_standard(iso834_time, 273.15 + 20)

    df_mc['case_name'] = [dict_in['case_name']] * n_simulations
    df_mc['time_step'] = [dict_in['time_step']] * n_simulations
    df_mc['time_limiting'] = [dict_in['fire_tlim']] * n_simulations
    df_mc['window_height'] = [dict_in['room_window_height']] * n_simulations
    df_mc['window_width'] = [dict_in['room_window_width']] * n_simulations
    df_mc['room_breadth'] = [dict_in['room_breadth']] * n_simulations
    df_mc['room_depth'] = [dict_in['room_depth']] * n_simulations
    df_mc['room_height'] = [dict_in['room_height']] * n_simulations
    df_mc['room_wall_thermal_inertia'] = [dict_in['room_wall_thermal_inertia']] * n_simulations
    df_mc['fire_duration'] = [dict_in['time_duration']] * n_simulations
    df_mc['fire_mode'] = [dict_in['fire_mode']] * n_simulations
    df_mc['fire_t_alpha'] = [dict_in['fire_t_alpha']] * n_simulations
    df_mc['fire_gamma_fi_q'] = [dict_in['fire_gamma_fi_q']] * n_simulations
    df_mc['beam_c'] = [beam_c] * n_simulations
    df_mc['beam_cross_section_area'] = [dict_in['beam_cross_section_area']] * n_simulations
    df_mc['beam_temperature_goal'] = [dict_in['beam_temperature_goal']] * n_simulations
    df_mc['beam_loc_z'] = [dict_in['beam_loc_z']] * n_simulations
    df_mc['protection_k'] = [dict_in['beam_protection_k']] * n_simulations
    df_mc['protection_rho'] = [dict_in['beam_protection_rho']] * n_simulations
    df_mc['protection_c'] = [dict_in['beam_protection_c']] * n_simulations
    df_mc['protection_thickness'] = [dict_in['beam_protection_thickness']] * n_simulations
    df_mc['protection_protected_perimeter'] = [dict_in['beam_protection_protected_perimeter']] * n_simulations
    df_mc['iso834_time'] = [iso834_time] * n_simulations
    df_mc['iso834_temperature'] = [iso834_temperature] * n_simulations
    df_mc['probability_weight'] = [dict_in['probability_weight'] / n_simulations] * n_simulations
    df_mc['index'] = np.arange(0, n_simulations, 1)

    return df_mc


def main(path_input_master: str = None):

    # ==================================================================================================================
    # Parse inputs from files into list of dict()s
    # ==================================================================================================================

    if path_input_master is None:
        path_input_master = select_path_input_csv()
    path_input_master = os.path.realpath(path_input_master)
    path_work = os.path.dirname(path_input_master)

    if path_input_master.endswith('.csv'):
        df_input_params = pd.read_csv(path_input_master).set_index('PARAMETERS')
    elif path_input_master.endswith('.xlsx'):
        df_input_params = pd.read_excel(path_input_master).set_index('PARAMETERS')
    else:
        df_input_params = []
    dict_dict_input_params = df_input_params.to_dict()

    # load configuration parameters
    try:
        with open(os.path.join(path_work, 'config.json'), 'r') as f:
            dict_config_params = json.load(f)
    except (FileExistsError, FileNotFoundError) as _:
        dict_config_params_default = {
            'n_threads': 1,
            'output_fires': 0,
        }
        dict_config_params = dict_config_params_default

    # ==================================================================================================================
    # Main calculation
    # ==================================================================================================================

    dict_out = main_params(input_master=dict_dict_input_params, config_master=dict_config_params)

    for file_name, df_ in dict_out.items():
        if isinstance(df_, pd.DataFrame):
            df_.to_csv(os.path.join(path_work, file_name), index=False)
        else:
            warnings.warn('DataFrame object is expected, got {}.'.format(type(df_)))


def main_params(input_master: dict = None, config_master: dict = None):
    """DESCRIPTION
    Run Monte Carlo Simulation for mc0 (time equivalence approach 0), with inputs and outputs as variables, no system IO
    is involved.

    PARAMETERS
    :param input_master:    {case_name: {arg1: val, arg2: val, ...}}, where 'case_name' is case name (i.e. input csv
                            file's column header and arg# are stochastic/static parameters of the case.
    :param config_master:   dict, configuration definition.
    :return dict_out:       dict(DataFrame), concatenated output, mcs results and fires (if enquired).
    """

    # ==================================================================================================================
    # Parse configurations
    # ==================================================================================================================

    if config_master:
        n_threads = 1 if 'n_threads' not in config_master else config_master['n_threads']
        output_fires = 1 if 'output_fires' not in config_master else config_master['output_fires']
        output_summary = 1 if 'output_summary' not in config_master else config_master['output_summary']
    else:
        n_threads = 1
        output_fires = 0
        output_summary = 1

    # ==================================================================================================================
    # Spawn Monte Carlo parameters from dict() to DataFrame()s
    # ==================================================================================================================
    
    list_df_mc = []
    for case_name, params in input_master.items():
        params['case_name'] = case_name
        list_df_mc.append(mc_inputs_dict_to_df(params))
    df_mc_params = pd.concat(list_df_mc)

    # ==================================================================================================================
    # Main calculation
    # ==================================================================================================================

    list_mcs_out = list()
    dict_out = dict()

    for case_name in input_master.keys():

        df_mc_params_i = df_mc_params[df_mc_params['case_name']==case_name]

        list_dict_mc_params_i = df_mc_params_i.to_dict('records')

        print('{:<24.24}: {}'.format("CASE", case_name))
        print('{:<24.24}: {}'.format("NO. OF THREADS", n_threads))
        print('{:<24.24}: {}'.format("NO. OF SIMULATIONS", len(df_mc_params_i.index)))

        if n_threads == 1:
            results = [calc_time_equivalence(**dict_mc_params) for dict_mc_params in tqdm(list_dict_mc_params_i, ncols=60)]

        else:
            m = mp.Manager()
            q = m.Queue()
            p = mp.Pool(n_threads, maxtasksperchild=1000)
            jobs = p.map_async(calc_time_equivalence_worker, [(dict_, q) for dict_ in list_dict_mc_params_i])
            count_total_simulations = len(list_dict_mc_params_i)

            with tqdm(total=count_total_simulations, ncols=60) as pbar:
                while True:
                    if jobs.ready():
                        if count_total_simulations > pbar.n:
                            pbar.update(count_total_simulations - pbar.n)
                        break
                    else:
                        if q.qsize() - pbar.n > 0:
                            pbar.update(q.qsize() - pbar.n)
                        time.sleep(1)
                p.close()
                p.join()

            results = jobs.get()

        df_output = pd.DataFrame(results)
        df_output.sort_values('solver_time_equivalence_solved', inplace=True)  # sort base on time equivalence

        # export fires

        fire_time = df_output['fire_time']
        fire_temperature = df_output['fire_temperature']
        if output_fires:
            list_fire_temperature = fire_temperature.values
            dict_fire_temperature = {
                'temperature_{}'.format(v): list_fire_temperature[i] for i, v in enumerate(df_output['index'].values)
            }
            dict_fire_temperature['time'] = fire_time.iloc[0]
            df_fire_temperature = pd.DataFrame.from_dict(dict_fire_temperature)
            dict_out['.'.join([case_name, 'fires'])] = df_fire_temperature

        if output_summary:
            time.sleep(0.5)
            print(y_results_summary(df_output))

        df_output.pop('fire_time')
        df_output.pop('fire_temperature')

        list_mcs_out.append(df_output)

    dict_out['mcs.out'] = pd.concat(list_mcs_out)

    return dict_out


def _test_mc_params_1():
    _input_master_ = dict(
        example_case_1=dict(
            is_live=1,
            fire_mode=3,
            probability_weight=1,
            n_simulations=1000,
            time_step=30,
            time_duration=18000,
            fire_hrr_density_lbound=0.25,
            fire_hrr_density_ubound=0.25,
            fire_hrr_density_mean=0.25,
            fire_hrr_density_std=100,
            fire_qfd_std=126,
            fire_qfd_mean=420,
            fire_qfd_ubound=1500,
            fire_qfd_lbound=10,
            fire_spread_lbound=0.0035,
            fire_spread_ubound=0.019,
            fire_nft_mean=1050,
            fire_com_eff_lbound=0.75,
            fire_com_eff_ubound=0.999,
            fire_tlim=0.333,
            fire_t_alpha=300,
            fire_gamma_fi_q=1,
            room_breadth=16,
            room_depth=31.25,
            room_height=3,
            room_window_width=72,
            room_window_height=2.8,
            room_opening_fraction_std=0.2,
            room_opening_fraction_mean=0.2,
            room_opening_fraction_ubound=0.999,
            room_opening_fraction_lbound=0.001,
            room_opening_permanent_fraction=0,
            room_wall_thermal_inertia=720,
            beam_cross_section_area=0.017,
            beam_rho=7850,
            beam_temperature_goal=893,
            beam_protection_protected_perimeter=2.14,
            beam_protection_thickness=0,
            beam_protection_k=0.2,
            beam_protection_rho=800,
            beam_protection_c=1700,
            beam_loc_z=3,
            beam_loc_ratio_lbound=0.666,
            beam_loc_ratio_ubound=0.999,
        ),
        example_case_2 = dict(
            is_live=1,
            fire_mode=3,
            probability_weight=1,
            n_simulations=1000,
            time_step=30,
            time_duration=18000,
            fire_hrr_density_lbound=0.75,
            fire_hrr_density_ubound=0.75,
            fire_hrr_density_mean=0.75,
            fire_hrr_density_std=100,
            fire_qfd_std=126,
            fire_qfd_mean=420,
            fire_qfd_ubound=1500,
            fire_qfd_lbound=10,
            fire_spread_lbound=0.0035,
            fire_spread_ubound=0.019,
            fire_nft_mean=1050,
            fire_com_eff_lbound=0.75,
            fire_com_eff_ubound=0.999,
            fire_tlim=0.333,
            fire_t_alpha=300,
            fire_gamma_fi_q=1,
            room_breadth=16,
            room_depth=31.25,
            room_height=3,
            room_window_width=72,
            room_window_height=2.8,
            room_opening_fraction_std=0.2,
            room_opening_fraction_mean=0.2,
            room_opening_fraction_ubound=0.999,
            room_opening_fraction_lbound=0.001,
            room_opening_permanent_fraction=0,
            room_wall_thermal_inertia=720,
            beam_cross_section_area=0.017,
            beam_rho=7850,
            beam_temperature_goal=893,
            beam_protection_protected_perimeter=2.14,
            beam_protection_thickness=0,
            beam_protection_k=0.2,
            beam_protection_rho=800,
            beam_protection_c=1700,
            beam_loc_z=3,
            beam_loc_ratio_lbound=0.666,
            beam_loc_ratio_ubound=0.999,
        )
    )

    _config_master_ = dict(
        n_threads=6,
        output_fires=0,
        output_summary=1,
    )

    main_params(input_master=_input_master_, config_master=_config_master_)


if __name__ == '__main__':

    warnings.filterwarnings('ignore')

    main()



if __name__ == '__main__':
    import sys
    import warnings
    import sfeprapy.mc1.mc1 as mc1

    warnings.filterwarnings('ignore')

    if len(sys.argv) > 1:
        mc1.main(sys.argv[1])
    else:
        mc1.main()



example_input = dict(
    example_case_1=dict(
        is_live=1,
        fire_mode=3,
        probability_weight=1,
        n_simulations=100,
        time_step=30,
        time_duration=18000,
        fire_hrr_density_lbound=0.25,
        fire_hrr_density_ubound=0.25,
        fire_hrr_density_mean=0.25,
        fire_hrr_density_std=100,
        fire_qfd_std=126,
        fire_qfd_mean=420,
        fire_qfd_ubound=1500,
        fire_qfd_lbound=10,
        fire_spread_lbound=0.0035,
        fire_spread_ubound=0.019,
        fire_nft_mean=1050,
        fire_com_eff_lbound=0.75,
        fire_com_eff_ubound=0.999,
        fire_tlim=0.333,
        fire_t_alpha=300,
        fire_gamma_fi_q=1,
        room_breadth=16,
        room_depth=31.25,
        room_height=3,
        room_window_width=72,
        room_window_height=2.8,
        room_opening_fraction_std=0.2,
        room_opening_fraction_mean=0.2,
        room_opening_fraction_ubound=0.999,
        room_opening_fraction_lbound=0.001,
        room_opening_permanent_fraction=0,
        room_wall_thermal_inertia=720,
        beam_cross_section_area=0.017,
        beam_rho=7850,
        beam_temperature_goal=893,
        beam_protection_protected_perimeter=2.14,
        beam_protection_thickness=0,
        beam_protection_k=0.2,
        beam_protection_rho=800,
        beam_protection_c=1700,
        beam_loc_z=3,
        beam_loc_ratio_lbound=0.666,
        beam_loc_ratio_ubound=0.999,
    ),
    example_case_2=dict(
        is_live=1,
        fire_mode=3,
        probability_weight=1,
        n_simulations=100,
        time_step=30,
        time_duration=18000,
        fire_hrr_density_lbound=0.75,
        fire_hrr_density_ubound=0.75,
        fire_hrr_density_mean=0.75,
        fire_hrr_density_std=100,
        fire_qfd_std=126,
        fire_qfd_mean=420,
        fire_qfd_ubound=1500,
        fire_qfd_lbound=10,
        fire_spread_lbound=0.0035,
        fire_spread_ubound=0.019,
        fire_nft_mean=1050,
        fire_com_eff_lbound=0.75,
        fire_com_eff_ubound=0.999,
        fire_tlim=0.333,
        fire_t_alpha=300,
        fire_gamma_fi_q=1,
        room_breadth=16,
        room_depth=31.25,
        room_height=3,
        room_window_width=72,
        room_window_height=2.8,
        room_opening_fraction_std=0.2,
        room_opening_fraction_mean=0.2,
        room_opening_fraction_ubound=0.999,
        room_opening_fraction_lbound=0.001,
        room_opening_permanent_fraction=0,
        room_wall_thermal_inertia=720,
        beam_cross_section_area=0.017,
        beam_rho=7850,
        beam_temperature_goal=893,
        beam_protection_protected_perimeter=2.14,
        beam_protection_thickness=0,
        beam_protection_k=0.2,
        beam_protection_rho=800,
        beam_protection_c=1700,
        beam_loc_z=3,
        beam_loc_ratio_lbound=0.666,
        beam_loc_ratio_ubound=0.999,
    )
)

EXAMPLE_CONFIG_DICT = dict(
    n_threads=2,
)

EXAMPLE_INPUT_DICT = {
    "Standard Case 1": dict(
        case_name='Standard Case 1',
        n_simulations=100,
        probability_weight=0.5,
        fire_time_step=30,
        fire_time_duration=18000,

        beam_position_horizontal=dict(
            dist='uniform_',
            lbound=31.25 * 0.666,
            ubound=31.25 * 0.999),
        fire_hrr_density=dict(
            dist='uniform_',
            lbound=0.25 - 0.001,
            ubound=0.25 + 0.001),
        fire_load_density=dict(
            dist='gumbel_r_',
            lbound=10,
            ubound=1500,
            mean=420,
            sd=126),
        fire_spread_speed=dict(
            dist='uniform_',
            lbound=0.0035,
            ubound=0.0190),
        fire_nft_limit=dict(
            dist='norm_',
            lbound=623.15,
            ubound=2023.15,
            mean=1323.15,
            sd=93),
        fire_combustion_efficiency=dict(
            dist='uniform_',
            lbound=0.8,
            ubound=1.0),
        window_open_fraction=dict(
            dist='lognorm_mod_',
            ubound=0.9999,
            lbound=0.0001,
            mean=0.2,
            sd=0.2),

        beam_cross_section_area=0.017,
        beam_position_vertical=3.2,
        beam_rho=7850,
        fire_mode=3.0,
        fire_gamma_fi_q=1,
        fire_t_alpha=300,
        fire_tlim=0.333,
        protection_c=1700,
        protection_k=0.2,
        protection_protected_perimeter=2.14,
        protection_rho=800,
        room_breadth=16,
        room_depth=31.25,
        room_height=3.3,
        room_wall_thermal_inertia=720,
        solver_temperature_goal=620 + 273.15,
        solver_max_iter=20,
        solver_thickness_lbound=0.0001,
        solver_thickness_ubound=0.0500,
        solver_tol=1.,
        window_height=2.8,
        window_width=72
    ),
    "Standard Case 2": dict(
        case_name='Standard Case 2',
        n_simulations=100,
        probability_weight=0.5,
        fire_time_step=30,
        fire_time_duration=18000,

        beam_position_horizontal=dict(
            dist='uniform_',
            lbound=31.25 * 0.666,
            ubound=31.25 * 0.999),
        fire_hrr_density=dict(
            dist='uniform_',
            lbound=0.25 - 0.001,
            ubound=0.25 + 0.001),
        fire_load_density=dict(
            dist='gumbel_r_',
            lbound=10,
            ubound=1500,
            mean=420,
            sd=126),
        fire_spread_speed=dict(
            dist='uniform_',
            lbound=0.0035,
            ubound=0.0190),
        fire_nft_limit=dict(
            dist='norm_',
            lbound=623.15,
            ubound=2023.15,
            mean=1323.15,
            sd=93),
        fire_combustion_efficiency=dict(
            dist='uniform_',
            lbound=0.8,
            ubound=1.0),
        window_open_fraction=dict(
            dist='lognorm_mod_',
            ubound=0.9999,
            lbound=0.0001,
            mean=0.2,
            sd=0.2),

        beam_cross_section_area=0.017,
        beam_position_vertical=3.2,
        beam_rho=7850,
        fire_mode=3.0,
        fire_gamma_fi_q=1,
        fire_t_alpha=300,
        fire_tlim=0.333,
        protection_c=1700,
        protection_k=0.2,
        protection_protected_perimeter=2.14,
        protection_rho=800,
        room_breadth=16,
        room_depth=31.25,
        room_height=3.3,
        room_wall_thermal_inertia=720,
        solver_temperature_goal=620 + 273.15,
        solver_max_iter=20,
        solver_thickness_lbound=0.0001,
        solver_thickness_ubound=0.0500,
        solver_tol=1.,
        window_height=2.8,
        window_width=72
    )
}

EXAMPLE_INPUT_CSV = """PARAMETERS, Standard Case 1, Standard Case 2
case_name,Standard Case 1,Standard Case 2
n_simulations,1000,1000
probability_weight,0.5,0.5
fire_time_step,30,30
fire_time_duration,18000,18000
beam_position_horizontal:dist,uniform_,uniform_
beam_position_horizontal:lbound,20.8125,20.8125
beam_position_horizontal:ubound,31.21875,31.21875
fire_hrr_density:dist,uniform_,uniform_
fire_hrr_density:lbound,0.26,0.26
fire_hrr_density:ubound,0.24,0.24
fire_load_density:dist,gumbel_r_,gumbel_r_
fire_load_density:lbound,10,10
fire_load_density:ubound,1500,1500
fire_load_density:mean,420,420
fire_load_density:sd,126,126
fire_spread_speed:dist,uniform_,uniform_
fire_spread_speed:lbound,0.0035,0.0035
fire_spread_speed:ubound,0.019,0.019
fire_nft_limit:dist,norm_,norm_
fire_nft_limit:lbound,623.15,623.15
fire_nft_limit:ubound,2023.15,2023.15
fire_nft_limit:mean,1323.15,1323.15
fire_nft_limit:sd,93,93
fire_combustion_efficiency:dist,uniform_,uniform_
fire_combustion_efficiency:lbound,0.8,0.8
fire_combustion_efficiency:ubound,1,1
window_open_fraction:dist,lognorm_mod_,lognorm_mod_
window_open_fraction:ubound,0.9999,0.9999
window_open_fraction:lbound,0.0001,0.0001
window_open_fraction:mean,0.2,0.2
window_open_fraction:sd,0.2,0.2
beam_cross_section_area,0.017,0.017
beam_position_vertical,3.2,3.2
beam_rho,7850,7850
fire_mode,3,3
fire_gamma_fi_q,1,1
fire_t_alpha,300,300
fire_tlim,0.333,0.333
protection_c,1700,1700
protection_k,0.2,0.2
protection_protected_perimeter,2.14,2.14
protection_rho,800,800
room_breadth,16,16
room_depth,31.25,31.25
room_height,3.3,3.3
room_wall_thermal_inertia,720,720
solver_temperature_goal,893.15,893.15
solver_max_iter,20,20
solver_thickness_lbound,0.0001,0.0001
solver_thickness_ubound,0.05,0.05
solver_tol,1,1
window_height,2.8,2.8
window_width,72,72
"""



# -*- coding: utf-8 -*-

import warnings
import numpy as np
import copy


def heat_transfer_general_1d_finite_difference(
        dt: float,
        k: np.ndarray,
        rho: np.ndarray,
        c: np.ndarray,
        T: np.ndarray,
        dx: np.ndarray,
        h0: float,
        T0: float,
        h1: float,
        T1: float
):
    """
    Calculates the current temperature for layers based on given thermal properties (i.e. k, rho, c, T and dx).
    This function returns a ndarray which contains temperatures of different layers, i.e. [T_0, T_1, T_2, ... T_n].

    NOTE:
    inert boundary - make inner_conductivity = 0.5 * dx[0] / k[0] or outer_conductivity = 0.5 * dx[n] / k[n]

    :param k:   [W/m/K]     thermal conductivity
    :param rho: [kg/m3]     density of boundary material
    :param c:   [J/kg/K]    specific heat of boundary material
    :param T:   [K]         temperature of boundary material (previous time step)
    :param dx:  [m]         virtual layers' thickness of boundary layer
    :param dt:  [s]         time step
    :param h0:  [W/m/K]     overall dQ/dt at j = -1 side (interior)
    :param T0:  [K]         temperature at j = -1 side (interior)
    :param h1:  [W/m/K]     overall dQ/dt at j = n+1 side (exterior)
    :param T1:  [K]         temperature of at j = n+1 side (exterior)
    :return:
    """
    # =================================================== PERMEABLE ====================================================
    # refactor parameters for readability

    k_ = 0.5 * (k[:1] + k[:-1])
    rho_ = 0.5 * (rho[1:] + rho[:-1])
    c_ = 0.5 * (c[1:] + c[:-1])

    T = T
    x_ = 0.5 * (dx[1:] + dx[:-1])
    h0 = h0
    T0 = T0
    h1 = h1
    T1 = T1

    # instantiate returning parameter T_i, current temperature at all virtual layers
    dT = np.zeros_like(T, dtype=float)

    # ====================================== FINITE DIFFERENCE METHOD CALCULATION ======================================

    j = 0
    left = h0 / rho[j] / c[j] / dx[j] * (T0 - T[j])
    right = k_[j] / rho_[j] / c_[j] / x_[j] * (T[j] - T[j+1])
    dT[j] = (left - right) * dt
    j = -1
    left = k_[j] / rho_[j] / c_[j] / x_[j] * (T[j] - T[j+1])
    right = h1 / rho[j] / c[j] / dx[j] * (T[j] - T1)
    dT[j] = (left - right) * dt

    print(dT[0], dT[-1])

    T[0] += dT[0]
    T[-1] += dT[-1]

    # left = k_[:-1] / rho_[:-1] / c_[:-1] / x_[:-1] * (T[:-2] - T[1:-1])
    # right = k_[1:] / rho_[1:] / c_[1:] / x_[1:] * (T[1:-1] - T[2:])
    # dTdt[1:-1] = left - right

    left = k_[:-1] / rho_[:-1] / c_[:-1] / (x_[:-1] ** 2) * (T[:-2] - T[1:-1])
    right = k_[1:] / rho_[1:] / c_[1:] / (x_[1:] ** 2) * (T[1:-1] - T[2:])
    dT[1:-1] = left - right

    return dT


def c_steel_T(temperature):

    temperature -= 273.15

    if temperature < 20:
        warnings.warn('Temperature ({:.1f} °C) is below 20 °C'.format(temperature))
        return 425 + 0.773 * 20 - 1.69e-3 * np.power(20, 2) + 2.22e-6 * np.power(20, 3)
    if 20 <= temperature < 600:
        return 425 + 0.773 * temperature - 1.69e-3 * np.power(temperature, 2) + 2.22e-6 * np.power(temperature, 3)
    elif 600 <= temperature < 735:
        return 666 + 13002 / (738 - temperature)
    elif 735 <= temperature < 900:
        return 545 + 17820 / (temperature - 731)
    elif 900 <= temperature <= 1200:
        return 650
    elif temperature > 1200:
        warnings.warn('Temperature ({:.1f} °C) is greater than 1200 °C'.format(temperature))
        return 650
    else:
        warnings.warn('Temperature ({:.1f} °C) is outside bound.'.format(temperature))
        return 0

from typing import Union


def steel_c_T_vectorised(T: np.ndarray):

    T = T - 273.15  # unit conversion: K -> deg.C

    c = np.where(T < 20, 425, 0)
    c = np.where((20 <= T) & (T <= 600), 425 + 0.773 * T - 1.69e-3 * np.power(T, 2) + 2.22e-6 * np.power(T, 3), c)
    c = np.where((600 < T) & (T <= 735), 666 + 13002 / (738 - T), c)
    c = np.where((735 < T) & (T <= 900), 545 + 17820 / (T - 731), c)
    c = np.where((900 < T) & (T <= 1200), 650, c)
    c = np.where(1200 < T, 651, c)

    return c


def k_steel_T(temperature):
    temperature -= 273.15
    if temperature < 20:
        warnings.warn('Temperature ({:.1f} °C) is below 20 °C'.format(temperature))
        return 54 - 3.33e-2 * 20
    if temperature < 800:
        return 54 - 3.33e-2 * temperature
    elif temperature <= 1200:
        return 27.3
    elif temperature > 1200:
        warnings.warn('Temperature ({:.1f} °C) is greater than 1200 °C'.format(temperature))
        return 27.3
    else:
        warnings.warn('Temperature ({:.1f} °C) is outside bound.'.format(temperature))
        return 0


def steel_k_T_vectorised(T: np.ndarray):

    T = T - 273.15  # unit conversion: K -> deg.C

    k = np.where(T < 20, 54, 0)
    k = np.where((20 <= T) & (T <= 800), 54 - 3.33e-2 * T, k)
    k = np.where((800 < T) & (T <= 1200), 27.3, k)
    k = np.where(1200 < T, 27.3, k)

    return k


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import seaborn as sns

    # solid properties
    thickness = 0.1
    m = 20
    rho = 7850  # kg/m3
    c_func = c_steel_T  # J/kg/K
    k_func = k_steel_T  # W/K/m

    # boundary properties
    U0 = 1000 + 273.15
    h0 = 25
    U1 = 20 + 273.15
    h1 = 4
    emissivity = 0.7

    # time properties
    t0 = 0
    t1 = 600

    # construct vectors
    thickness += (thickness / m * 2)

    m += 2

    x = np.linspace(0, thickness, 2*m+1)[1::2]  # location of nodes

    dx = np.full((m,), thickness/m)  # thickness of nodes
    dx[[0, -1]] = 1  # gas nodes have unity thickness

    U = np.ones_like(dx) * (20 + 273.15)  # temperature of nodes

    rho = np.full_like(dx, rho)

    h = np.zeros_like(dx)  # convection coefficients
    h[0], h[-1] = h0, h1  # convection at gas nodes

    e = np.zeros_like(dx)  # radiation heat transfer coefficient
    e[[0, -1]] = 5.67e-8 * emissivity  # radiation at gas nodes

    # ==================================================================================================================
    # ONE-DIMENSIONAL FINITE DIFFERENCE START
    # ==================================================================================================================

    c = np.full_like(dx, 460)
    # k = np.full_like(dx, 30)

    dU_dt_ = list()  # results container - temperature change rate
    U_ = list()

    t = t0  # set simulation time
    while t <= t1:
        print(int(t))

        if int(t) == 300:
            print('d')

        c = steel_c_T_vectorised(U)
        # k = np.fromiter(map(k_steel_T, U), dtype=float) / dx
        k = steel_k_T_vectorised(U) / dx
        k[[0, -1]] = 0

        dt = 0.005  # constant time step

        # Temperatures at boundary (i.e. gas)
        # ===================================

        U[0], U[-1] = U0, U1  # boundary at constant temperature
        # Temperatures at (solid) outer nodes, only energy in and out, no into other solids
        # =================================================================================
        u0 = U[0] - U[1]
        u1 = U[1] - U[2]
        u0_4 = U[0]**4 - U[1]**4
        u1_4 = U[1]**4 - U[2]**4
        e0 = u0 * h[0] - u1 * h[1]  # convection
        e1 = u0 * k[0] - u1 * k[1]  # conduction
        e2 = u0_4 * e[0] - u1_4 * e[1]  # radiation
        e3 = rho[1] * c[1] * dx[1]  # stored
        dU_dt_0 = (e0 + e1 + e2) / e3
        U[1] = U[1] + dU_dt_0 * dt

        # Temperatures at intermediate nodes
        # ==================================

        for m in range(2, len(dx)-2, 1):
            du0 = U[m-1] - U[m]
            du1 = U[m] - U[m+1]
            e1 = du0 * k[m] - du1 * k[m]
            en = rho[m] * c[m] * dx[m]
            du_dt = e1 / en
            U[m] += du_dt * dt

        # u0 = U[1:-3] - U[2:-2]
        # u1 = U[2:-2] - U[3:-1]
        # e0 = u0 * h[1:-3] - u1 * h[3:-1]
        # e1 = u0 * k[2:-2] - u1 * k[2:-2]
        # e3 = rho[2:-2] * c[2:-2] * dx[2:-2] ** 2
        # dU_dt = (e0 + e1) / e3
        # U[2:-2] = U[2:-2] + dU_dt * dt

        # Temperatures at (solid) outer nodes, only energy in and out, no into other solids
        # =================================================================================

        u0 = U[-3] - U[-2]
        u1 = U[-2] - U[-1]
        u0_4 = U[-3]**4 - U[-2]**4
        u1_4 = U[-2]**4 - U[-1]**4
        e0 = u0 * h[-2] - u1 * h[-1]  # convection
        e1 = u0 * k[-2] - u1 * k[-1]  # conduction
        e2 = u0_4 * e[-2] - u1_4 * e[-1]  # radiation
        e3 = rho[-2] * c[-2] * dx[-2]  # stored
        dU_dt_1 = (e0 + e1 + e2) / e3
        U[-2] = U[-2] + dU_dt_1 * dt

        # dU_dt_.append(dU_dt)
        U_.append(copy.copy(U))

        # Increment time
        # ==============

        t += dt

    fig, ax1 = plt.subplots()
    # dU_dt_ = np.stack(dU_dt_, axis=0)
    # U_stacked = np.stack(U_, axis=0)
    # sns.heatmap(U_stacked[0::100, 1:-1], ax=ax1, xticklabels=np.round(x[1:-1], 2), yticklabels=[])
    for u in U_[::int(len(U_)/10)]:
        sns.lineplot(x[1:-1], u[1:-1]-273.15, ax=ax1)
    sns.lineplot(x[1:-1], U[1:-1]-273.15, ax=ax1, label=int(t))

    plt.tight_layout()
    plt.show()



# -*- coding: utf-8 -*-
import copy
from typing import Union

import numpy as np


def fire(
        t: np.array,
        fire_load_density_MJm2: float,
        fire_hrr_density_MWm2: float,
        room_length_m: float,
        room_width_m: float,
        fire_spread_rate_ms: float,
        beam_location_height_m: float,
        beam_location_length_m: Union[float, list, np.ndarray],
        fire_nft_limit_c: float,
        opening_fraction: float = 0,
        opening_width_m: float = 0,
        opening_height_m: float = 0,
):
    """
    This function calculates and returns a temperature array representing travelling fire. This function is NOT in SI.
    :param t: in s, is the time array
    :param fire_load_density_MJm2: in MJ/m2, is the fuel density on the floor
    :param fire_hrr_density_MWm2: in MW/m2, is the heat release rate density
    :param room_length_m: in m, is the room length
    :param room_width_m: in m, is the room width
    :param fire_spread_rate_ms: in m/s, is fire spread speed
    :param beam_location_height_m: in m, is the beam lateral distance to fire origin
    :param beam_location_length_m: in m, is the beam height above the floor
    :param fire_nft_limit_c: in deg.C, is the maximum near field temperature
    :param opening_fraction: in -, is the ventilation opening proportion between 0 to 1
    :param opening_width_m: in m, is ventilation opening width
    :param opening_height_m: in m, is ventilation opening height
    :return T_g: in deg.C, is calculated gas temperature
    """

    # re-assign variable names for equation readability
    q_fd = fire_load_density_MJm2
    HRRPUA = fire_hrr_density_MWm2
    l = max([room_length_m, room_width_m])
    w = min([room_length_m, room_width_m])
    s = fire_spread_rate_ms
    h_s = beam_location_height_m
    l_s = beam_location_length_m

    # work out ventilation conditions

    # a_v = opening_height_m * opening_width_m * opening_fraction
    # Qv = 1.75 * a_v * np.sqrt(opening_height_m)

    # workout burning time etc.
    t_burn = max([q_fd / HRRPUA, 900.])
    t_decay = max([t_burn, l / s])
    t_lim = min([t_burn, l / s])

    # reduce resolution to fit time step for t_burn, t_decay, t_lim
    time_interval_s = t[1] - t[0]
    t_decay_ = round(t_decay / time_interval_s, 0) * time_interval_s
    t_lim_ = round(t_lim / time_interval_s, 0) * time_interval_s
    if t_decay_ == t_lim_: t_lim_ -= time_interval_s

    # workout the heat release rate ARRAY (corrected with time)
    Q_growth = (HRRPUA * w * s * t) * (t < t_lim_)
    Q_peak = min([HRRPUA * w * s * t_burn, HRRPUA * w * l]) * (t >= t_lim_) * (t <= t_decay_)
    Q_decay = (max(Q_peak) - (t - t_decay_) * w * s * HRRPUA) * (t > t_decay_)
    Q_decay[Q_decay < 0] = 0
    Q = (Q_growth + Q_peak + Q_decay) * 1000.

    # workout the distance between fire median to the structural element r
    l_fire_front = s * t
    l_fire_front[l_fire_front < 0] = 0
    l_fire_front[l_fire_front > l] = l
    l_fire_end = s * (t - t_lim)
    l_fire_end[l_fire_end < 0] = 0.
    l_fire_end[l_fire_end > l] = l
    l_fire_median = (l_fire_front + l_fire_end) / 2.

    # workout the far field temperature of gas T_g
    if isinstance(l_s, float) or isinstance(l_s, int):
        r = np.absolute(l_s - l_fire_median)
        T_g = np.where((r / h_s) > 0.8, (5.38 * np.power(Q / r, 2 / 3) / h_s) + 20., 0)
        T_g = np.where((r / h_s) <= 0.8, (16.9 * np.power(Q, 2 / 3) / np.power(h_s, 5 / 3)) + 20., T_g)
        T_g[T_g >= fire_nft_limit_c] = fire_nft_limit_c
        return T_g
    elif isinstance(l_s, np.ndarray) or isinstance(l_s, list):
        l_s_list = copy.copy(l_s)
        T_g_list = list()
        for l_s in l_s_list:
            r = np.absolute(l_s - l_fire_median)
            T_g = np.where((r / h_s) > 0.8, (5.38 * np.power(Q / r, 2 / 3) / h_s) + 20., 0)
            T_g = np.where((r / h_s) <= 0.8, (16.9 * np.power(Q, 2 / 3) / np.power(h_s, 5 / 3)) + 20., T_g)
            T_g[T_g >= fire_nft_limit_c] = fire_nft_limit_c
            T_g_list.append(T_g)
        return T_g_list
    else:
        raise TypeError('Unknown type of parameter "l_s": {}'.format(type(l_s)))


def fire_backup(
        t: np.ndarray,
        T_0: float,
        q_fd: float,
        hrrpua: float,
        l: float,
        w: float,
        s: float,
        e_h: float,
        e_l: float,
        T_max: float = 1323.15,
):
    """
    :param t: ndarray, [s] An array representing time incorporating 'temperature'.
    :param T_0: float, [K] ,Initial temperature.
    :param q_fd: float, [J/m2], Fire load density.
    :param hrrpua: float, [W/m2], Heat release rate density.
    :param l: float, [m], Compartment length.
    :param w: float, [m], Compartment width.
    :param s: float, [m/s], Fire spread speed.
    :param e_h: float, [m], Vertical distance between element to fuel bed.
    :param e_l: float, [m], Horizontal distance between element to fire front.
    :return temperature: [K] An array representing temperature incorporating 'time'.
    """

    # UNIT CONVERSION TO FIT EQUATIONS
    T_0 -= 273.15
    q_fd /= 1e6
    hrrpua /= 1e6
    T_max -= 273.15

    # workout time step
    time_step = t[1] - t[0]

    # workout burning time etc.
    t_burn = max([q_fd / hrrpua, 900.])
    t_decay = max([t_burn, l / s])
    t_lim = min([t_burn, l / s])

    # reduce resolution to fit time step for t_burn, t_decay, t_lim
    t_decay_ = round(t_decay / time_step, 0) * time_step
    t_lim_ = round(t_lim / time_step, 0) * time_step
    if t_decay_ == t_lim_: t_lim_ -= time_step

    # workout the heat release rate ARRAY (corrected with time)
    Q_growth = (hrrpua * w * s * t) * (t < t_lim_)
    Q_peak = min([hrrpua * w * s * t_burn, hrrpua * w * l]) * (t >= t_lim_) * (t <= t_decay_)
    Q_decay = (max(Q_peak) - (t - t_decay_) * w * s * hrrpua) * (t > t_decay_)
    Q_decay[Q_decay < 0] = 0
    Q = (Q_growth + Q_peak + Q_decay) * 1000.

    # workout the distance between fire_curve median to the structural element r
    l_fire_front = s * t
    l_fire_front[l_fire_front < 0] = 0.
    l_fire_front[l_fire_front > l] = l
    l_fire_end = s * (t - t_lim)
    l_fire_end[l_fire_end < 0] = 0.
    l_fire_end[l_fire_end > l] = l
    l_fire_median = (l_fire_front + l_fire_end) / 2.
    r = np.absolute(e_l - l_fire_median)
    r[r == 0] = 0.001  # will cause crash if r = 0

    # workout the far field temperature of gas T_g
    T_g1 = (5.38 * np.power(Q / r, 2 / 3) / e_h) * ((r / e_h) > 0.18)
    T_g2 = (16.9 * np.power(Q, 2 / 3) / np.power(e_h, 5 / 3)) * ((r / e_h) <= 0.18)
    T_g = T_g1 + T_g2 + T_0

    T_g[T_g >= T_max] = T_max

    # UNIT CONVERSION TO FIT OUTPUT (SI)
    T_g = T_g + 273.15  # C -> K
    Q *= 10e6  # MJ -> J

    return T_g


def test():
    def test_fire_backup():
        import numpy as np

        time = np.arange(0, 22080, 30)
        list_l = [50, 100, 150]

        import matplotlib.pyplot as plt
        plt.style.use('seaborn-paper')
        fig, ax = plt.subplots(figsize=(3.94, 2.76))
        ax.set_xlabel('Time [minute]')
        ax.set_ylabel('Temperature [$^{\circ}C$]')

        for l in list_l:
            temperature = fire_backup(
                t=time,
                T_0=293.15,
                q_fd=900e6,
                hrrpua=0.15e6,
                l=l,
                w=17.4,
                s=0.012,
                e_h=3.5,
                e_l=l / 2,
            )
            ax.plot(time / 60, temperature - 273.15)

        ax.legend().set_visible(True)
        ax.set_xlim((0, 120))
        ax.grid(color='k', linestyle='--')
        plt.tight_layout()
        plt.show()

    def test_fire():
        time = np.arange(0, 210 * 60, 30)
        list_l = [25, 50, 100, 150]

        import matplotlib.pyplot as plt
        plt.style.use('seaborn-paper')
        fig, ax = plt.subplots(figsize=(3.94, 2.76))
        ax.set_xlabel('Time [minute]')
        ax.set_ylabel('Temperature [$^{\circ}C$]')

        for length in list_l:
            temperature = fire(
                t=time,
                fire_load_density_MJm2=600,
                fire_hrr_density_MWm2=0.25,
                room_length_m=length,
                room_width_m=16,
                fire_spread_rate_ms=0.012,
                beam_location_height_m=3,
                beam_location_length_m=length / 2,
                fire_nft_limit_c=1050,
            )

            ax.plot(time / 60, temperature, label="Room length {:4.0f} m".format(length))

        ax.legend(loc=4).set_visible(True)
        ax.set_xlim((-10, 190))
        ax.grid(color='k', linestyle='--')
        plt.tight_layout()
        plt.show()

    def test_fire_multiple_beam_location():
        time = np.arange(0, 210 * 60, 30)
        length = 100

        import matplotlib.pyplot as plt
        plt.style.use('seaborn-paper')
        fig, ax = plt.subplots(figsize=(3.94, 2.76))
        ax.set_xlabel('Time [minute]')
        ax.set_ylabel('Temperature [$^{\circ}C$]')

        temperature_list = fire(
            t=time,
            fire_load_density_MJm2=600,
            fire_hrr_density_MWm2=0.25,
            room_length_m=length,
            room_width_m=16,
            fire_spread_rate_ms=0.012,
            beam_location_height_m=3,
            beam_location_length_m=np.linspace(0, length, 12)[1:-1],
            fire_nft_limit_c=1050,
        )

        for temperature in temperature_list:
            ax.plot(time / 60, temperature, label="Room length {:4.0f} m".format(length))

        ax.legend(loc=4).set_visible(True)
        ax.set_xlim((-10, 190))
        ax.grid(color='k', linestyle='--')
        plt.tight_layout()
        plt.show()

    test_fire_multiple_beam_location()


if __name__ == '__main__':
    test()



#  Danny Hopkin
#  OFR Consultants UK
#  Zone model sub-functions
#  16/05/18

import numpy as np

#  calculate max HRR

def hrrmax(h0,b0):

    # Returns output in kW

    a0 = b0*h0
    hrrmax = 1500 * a0 * np.sqrt(h0)
    return hrrmax

#  convective flow from opening

def convf_open(h0,b0, Tgas, Tamb, cp):

    # Outcome depends on units of cp

    a0 = h0*b0
    dT = Tgas - Tamb
    qc = 0.5*a0*np.sqrt(h0)*cp*dT
    return qc

# radiative flow from opening

def radf_open(h0, b0, Tgas, Tamb, h):

    #  Outcome in watts
    Tgas = Tgas+273
    Tamb = Tamb+273
    ef = 1- np.exp(-1.1*h)
    dT4 = (Tgas**4)-(Tamb**4)
    a0 = b0*h0
    boltz = 5.67e-8
    qr = a0*ef*boltz*dT4
    return qr

# Temperature rise of gas

def delta_Tgas(rhoamb,cp, b,d,h, dt, qgas):

    # Rise in K

    Vg = b*d*h
    dTgas = (qgas*dt)/(rhoamb*Vg*cp)
    return dTgas

# Energy stored in gas

def qgas_calc(HRR,qw,qc,qr):

    #  All parameters to be dimensionally consistent

    qgas = HRR-qw-qc-qr
    return qgas

# Heat transfer function to update wall temperature

def wall_HT_sub(gas_temp, temp_arr, nodes, dt, dx, hchot, hccold,effemish,effemisc,tamb):

    from scipy.interpolate import interp1d

    # Set heat transfer properties based upon initial temperature conditions

    lamda_temp_arr = [0, 300, 700, 1400]             #  Temperature array for conductivity [DegC]
    lamda_input_arr =[0.12,0.2, 0.4, 1.2]         #  Corresponding conductivity input [W/m.K]
    lamda_interp = interp1d(lamda_temp_arr,lamda_input_arr)     #  Interpolation function

    cp_temp_arr = [0, 1400]                 #  Temperature array for specific heat [DegC]
    cp_input_arr =[1600,1600]              #  Corresponding specific heat input [J/kg.K]
    cp_interp = interp1d(cp_temp_arr,cp_input_arr)              #  Interpolation function

    rho_temp_arr = [0, 1400]                                    #  Temperature array for density [DegC]
    rho_input_arr =[1,1]                                    #  Corresponding density ratio input [-]
    rho_interp = interp1d(rho_temp_arr,rho_input_arr)           #  Interpolation function
    density = 500                                              #  Density in [kg/m3]

    #  Main heat transfer solver

    lamdaNL = lamda_interp(temp_arr)
    cpNL = cp_interp(temp_arr)
    rhoNL = rho_interp(temp_arr) * density

    # First element calculations

    Qinc = ONEDHT_QINC(gas_temp, temp_arr[0], effemish, hchot)
    temp_arr[0] = ONEDHT_ELEM1(Qinc, temp_arr[0], temp_arr[1], lamdaNL[0], lamdaNL[1], dx, dt, cpNL[0], rhoNL[0])

    # Intermediate element calculations

    for nodenum in range(1, nodes - 1):
        temp_arr[nodenum] = ONEDHT_ELEMJ(temp_arr[nodenum - 1], temp_arr[nodenum], temp_arr[nodenum + 1],
                                         lamdaNL[nodenum - 1], lamdaNL[nodenum], lamdaNL[nodenum + 1], dx, dt,
                                         cpNL[nodenum], rhoNL[nodenum])

    # Final element calculations

    Qout = ONEDHT_QOUT(temp_arr[nodes - 1], tamb, effemisc, hccold)
    temp_arr[nodes - 1] = ONEDHT_ELEMF(Qout, temp_arr[nodes - 2], temp_arr[nodes - 1], lamdaNL[nodes - 2],
                                       lamdaNL[nodes - 1], dx, dt, cpNL[nodes - 1], rhoNL[nodes - 1])

    Temps=temp_arr
    return Temps

if __name__ == '__main__':

    # Imports

    from scipy.interpolate import interp1d
    from HT_sub_functions import ONEDHT_QINC, ONEDHT_QOUT, ONEDHT_ELEM1, ONEDHT_ELEMJ, ONEDHT_ELEMF
    from complex_intersect import complex_intersect
    from matplotlib import pyplot as plt
    from crib_mass_loss_functions import fuel_contr_crib

    #   Define room parameters

    depth = 4.5       # [m]
    breadth = 3.5     # [m]
    height = 2.5    # [m]

    #   Define opening parameters

    op_height = 2   #  [m]
    op_width = 1    #  [m]

    #   max HRR

    maxhrr = hrrmax(op_height,op_width)*1000    # in [W]

    # CLT related parameters

    CLT_mass = 250    #  initialise crib mass
    mass_lost = 0
    crib_mass_remain = 0
    delta_mass = 0  #  initialise change in crib mass
    crib_MLR = 0    #  initialise crib MLR
    crib_mass_consumed = 0  # initialise mass consumed term
    GER = 2                 # Global equivalence ratio

    #   calculate the area of lining

    Aopen = op_height * op_width
    Afloor = depth * breadth
    Aceiling = Afloor
    Awalls = 2*(depth + breadth) * height
    Alining = Afloor + Aceiling + Awalls - Aopen

    #   Define ambient conditions

    Tambient = 18       #  [C]
    Twall = 20          #  [C]
    Tgas = Tambient     #  set initial gas temperature to ambient

    #   Design fire input

    hrr_temp_arr = [0,300, 600, 1200, 1800, 3600, 12000]                   # Time array [s]
    hrr_input_arr =[0,100, 5000, 4000, 1200, 500, 0]    # HRR array [kW]
    hrr_interp = interp1d(hrr_temp_arr,hrr_input_arr)                  # Interpolation function

    # Time parameters

    uptime = 10000   #  [s]
    dt = 1          #  [s]
    endtimestep = int(uptime / dt)
    nth = 250

    #  Set initial wall conditions & properties
    nodes = 101  # set the number of nodes
    dx = 0.002
    wall_depth = dx*(nodes-1)
    temp_arr = np.ones(nodes)*20    # creates initial temperature nodal array with ambient condition
    wall_depth_arr = np.ones(nodes) # creates an array for interpolating an isotherm position

    # Generate depth array [mm]
    for x in range(0,nodes):
        wall_depth_arr[x]=x*dx*1000  # expressed in mm

    # Wall properties and isotherm tracking
    hchot = 25
    hccold = 9
    emissh = 0.7
    emissc = 0.7
    iso = 300
    iso_pos = 0

    #  Initialise complex intersection for isotherm tracking

    f = np.ones(nodes)*iso  # Creates a straight line with same number of indices for isotherm tracking

    # Create storage arrays

    time_array = []
    crib_mlr_arr = []
    wall_surf_T_array =[]
    Tgas_array = []
    isotherm_arr = []
    isotherm_arr2 = []

    total_mass_lost = 0
    mlr_max_crib = 0

    for tstep in range(1, endtimestep + 1):

        #  calculate time and corresponding HRR from input
        time = tstep * dt

        hrr = hrr_interp(time)*1000     # Convert to watts

        if hrr > maxhrr:
            hrr = maxhrr

        # Gas properties
        gas_temp_arr = [0, 230, 1300]  # Temp array
        cp_input_arr = [1003, 1030, 1216]  # Specific heat array
        cp_interp = interp1d(gas_temp_arr, cp_input_arr)

        # calculate wall losses
        qwall = ONEDHT_QINC(Tgas,Twall,emissh,hchot)
        qwall_tot = qwall * Alining

        # calculate opening losses
        qcon = convf_open(op_height,op_width,Tgas,0,cp_interp(Tgas))
        qrad =radf_open(op_height,op_width,Tgas,0,height)

        # calculate energy stored in the gas
        qgas = qgas_calc(hrr,qwall_tot,qcon,qrad)

        #  calculate rise in gas temperature
        dTgas = delta_Tgas(1,cp_interp(Tgas),breadth,depth,height,dt,qgas)

        #  update gas temperature
        Tgas = min(Tgas + dTgas,1300)

        # HT to wall

        temp_arr = wall_HT_sub(Tgas,temp_arr,nodes,dt, dx,hchot,hccold,emissh,emissc,Tambient)
        Twall = temp_arr[0]

        print(qwall, Twall, Tgas)

        # Update ambient condition in wall
        for xx in range(0, nodes):
            if temp_arr[xx] < 20:
                temp_arr[xx] = 20

        # Track isotherm
        idx = complex_intersect(f, temp_arr)
        iso_old = iso_pos

        if max(temp_arr) < iso:
            iso_pos = 0
        elif min(temp_arr) > iso:
            iso_pos = wall_depth
        else:
            iso_pos = idx * dx * 1000
            iso_pos = max(iso_pos)

        if iso_old > iso_pos:
            iso_pos = iso_old
        isotherm_arr.append(float(iso_pos))

        iso_change = max(iso_pos - iso_old,0)

        if iso_change > 0:
            exposed_area = 4.5*2.5
            delta_mass = exposed_area * (iso_change/1000) * 450
            CLT_mass = CLT_mass + delta_mass

        #  Calculate crib mass loss for lining
        if CLT_mass > 0:
            crib_MLR = fuel_contr_crib(0.05,crib_mass_remain,CLT_mass)
            mass_lost = crib_MLR*dt
            total_mass_lost = total_mass_lost + mass_lost
            crib_mass_remain = max(CLT_mass - total_mass_lost,0)

        crib_mlr_arr.append(crib_MLR)

        time_array.append(time)
        Tgas_array.append(Tgas)
        wall_surf_T_array.append(temp_arr[0])

        if tstep/nth == int(tstep/nth):
            if tstep > 1:
                plt.subplot(221)
                plt.plot(wall_depth_arr, temp_arr, label=str(tstep*dt))

        print('Time = ', time, ' [s]', 'Isotherm depth =', iso_pos, ' [mm]', 'New CLT mass =', CLT_mass, ' [kg]')
        print('Mass lost', mass_lost, total_mass_lost, crib_mass_remain)

    plt.xlabel('Depth [mm]')
    plt.ylabel('Temperature [$^o$ C]')
    plt.grid(True)
    plt.legend()
    plt.subplot(222)
    plt.xlabel('Time [s]')
    plt.ylabel('Isotherm depth [mm]')
    plt.grid(True)
    plt.plot(time_array, isotherm_arr)
    plt.subplot(223)
    plt.xlabel('Time [s]')
    plt.ylabel('Gas temperature [$^o$ C]')
    plt.plot(time_array, Tgas_array)
    plt.grid(True)
    plt.subplot(224)
    plt.xlabel('Time [s]')
    plt.ylabel('CLT MLR [kg/s]')
    plt.plot(time_array, crib_mlr_arr)
    plt.grid(True)
    plt.show()


import numpy as np


def fire(
        time,
        temperature_initial
):
    time /= 1200.  # convert time from seconds to hours
    temperature_initial -= 273.15  # convert ambient temperature from kelvin to celsius
    temperature = 660 * (1 - 0.687 * np.exp(-0.32 * time) - 0.313 * np.exp(-3.8 * time)) + temperature_initial
    return temperature + 273.15  # convert temperature from celsius to kelvin



# Imports

from .HT_sub_functions import ONEDHT_QINC, ONEDHT_QOUT, ONEDHT_ELEM1, ONEDHT_ELEMJ, ONEDHT_ELEMF
import copy
from matplotlib import pyplot as plt
import numpy as np

def wall_HT_sub(gas_temp, temp_arr, nodes, dt, dx, hchot, hccold,effemish,effemisc,tamb):

    from scipy.interpolate import interp1d

    # Set heat transfer properties based upon initial temperature conditions

    lamda_temp_arr = [0, 300, 700, 1400]             #  Temperature array for conductivity [DegC]
    lamda_input_arr =[0.12,0.2, 0.4, 1.2]         #  Corresponding conductivity input [W/m.K]
    lamda_interp = interp1d(lamda_temp_arr,lamda_input_arr)     #  Interpolation function

    cp_temp_arr = [0, 1400]                 #  Temperature array for specific heat [DegC]
    cp_input_arr =[1600,1600]              #  Corresponding specific heat input [J/kg.K]
    cp_interp = interp1d(cp_temp_arr,cp_input_arr)              #  Interpolation function

    rho_temp_arr = [0, 1400]                                    #  Temperature array for density [DegC]
    rho_input_arr =[1,1]                                    #  Corresponding density ratio input [-]
    rho_interp = interp1d(rho_temp_arr,rho_input_arr)           #  Interpolation function
    density = 500                                              #  Density in [kg/m3]

    #  Main heat transfer solver

    lamdaNL = lamda_interp(temp_arr)
    cpNL = cp_interp(temp_arr)
    rhoNL = rho_interp(temp_arr) * density

    # First element calculations

    Qinc = ONEDHT_QINC(gas_temp, temp_arr[0], effemish, hchot)
    temp_arr[0] = ONEDHT_ELEM1(Qinc, temp_arr[0], temp_arr[1], lamdaNL[0], lamdaNL[1], dx, dt, cpNL[0], rhoNL[0])

    # Intermediate element calculations

    for nodenum in range(1, nodes - 1):
        temp_arr[nodenum] = ONEDHT_ELEMJ(temp_arr[nodenum - 1], temp_arr[nodenum], temp_arr[nodenum + 1],
                                         lamdaNL[nodenum - 1], lamdaNL[nodenum], lamdaNL[nodenum + 1], dx, dt,
                                         cpNL[nodenum], rhoNL[nodenum])

    # Final element calculations

    Qout = ONEDHT_QOUT(temp_arr[nodes - 1], tamb, effemisc, hccold)
    temp_arr[nodes - 1] = ONEDHT_ELEMF(Qout, temp_arr[nodes - 2], temp_arr[nodes - 1], lamdaNL[nodes - 2],
                                       lamdaNL[nodes - 1], dx, dt, cpNL[nodes - 1], rhoNL[nodes - 1])

    Temps=temp_arr
    return Temps

if __name__ == '__main__':

    nodes = 11                      # set the number of nodes
    dx = 0.0025                     # in m
    temp_arr = np.ones(nodes)*20    # creates initial temperature nodal array with ambient condition
    endtime = 1800                  # number of timesteps
    dt = 1                          # time step increment - very important for numerical stability
    nth = 300                       # Temperature profile will be produced every nth time step
    endtimestep = int(endtime/dt)   # calculate total number of timesteps
    node_array =np.ones(nodes)      # Create depth array for temperature profile
    track_iso = 300                 # Isotherm depth to be tracked

    wall_depth_arr = np.ones(nodes) # creates an array for interpolating an isotherm position & plotting

    # Generate depth array [mm]
    for x in range(0,nodes):
        wall_depth_arr[x]=x*dx*1000  # expressed in mm

    list_temp_arr=[]

    # Main heat transfer solver

    for tstep in range(1, endtimestep + 1):

        gas_temp = 500
        temp_arr = wall_HT_sub(gas_temp,temp_arr,nodes,dt,dx,35,4,0.7,0.7,20)   # HT function

        # Store results

        list_temp_arr.append(copy.copy(temp_arr))

        if tstep/nth == int(tstep/nth):
            if tstep > 1:
                plt.plot(wall_depth_arr, temp_arr, label=str(tstep*dt))

        print(temp_arr[3])

    plt.xlabel('Depth [mm]')
    plt.ylabel('Temperature [$^o$ C]')
    plt.grid(True)
    plt.legend()
    plt.show()


# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import scipy.stats as stats


def gumbel_r_(mean: float, sd: float, **_):

    # parameters Gumbel W&S
    alpha = 1.282 / sd
    u = mean - 0.5772 / alpha

    # parameters Gumbel scipy
    scale = 1 / alpha
    loc = u

    return dict(loc=loc, scale=scale)


def lognorm_(mean: float, sd: float, **_):

    cov = sd / mean

    sigma_ln = np.sqrt(np.log(1 + cov ** 2))
    miu_ln = np.log(mean) - 1 / 2 * sigma_ln ** 2

    s = sigma_ln
    loc = 0
    scale = np.exp(miu_ln)

    return dict(s=s, loc=loc, scale=scale)


def norm_(mean: float, sd: float, **_):

    loc = mean
    scale = sd

    return dict(loc=loc, scale=scale)


def uniform_(ubound: float, lbound: float, **_):

    if lbound > ubound:
        lbound += ubound
        ubound = lbound - ubound
        lbound -= ubound

    loc = lbound
    scale = ubound - lbound

    return dict(loc=loc, scale=scale)


def random_variable_generator(dict_in: dict, num_samples: int):

    # assign distribution type
    dist_0 = dict_in['dist']
    dist = dict_in['dist']

    # assign distribution boundary (for samples)
    ubound = dict_in['ubound']
    lbound = dict_in['lbound']

    # sample CDF points (y-axis value)
    def generate_cfd_q(dist_, dist_kw_, lbound_, ubound_):
        cfd_q_ = np.linspace(
            getattr(stats, dist_).cdf(x=lbound_, **dist_kw_),
            getattr(stats, dist_).cdf(x=ubound_, **dist_kw_),
            num_samples)
        samples_ = getattr(stats, dist_).ppf(q=cfd_q_, **dist_kw_)
        return samples_

    # convert human distribution parameters to scipy distribution parameters
    if dist_0 == 'gumbel_r_':
        dist_kw = gumbel_r_(**dict_in)
        dist = 'gumbel_r'
        samples = generate_cfd_q(dist_=dist, dist_kw_=dist_kw, lbound_=lbound, ubound_=ubound)
    elif dist_0 == 'uniform_':
        dist_kw = uniform_(**dict_in)
        dist = 'uniform'
        samples = generate_cfd_q(dist_=dist, dist_kw_=dist_kw, lbound_=lbound, ubound_=ubound)

    elif dist_0 == 'norm_':
        dist_kw = norm_(**dict_in)
        dist = 'norm'
        samples = generate_cfd_q(dist_=dist, dist_kw_=dist_kw, lbound_=lbound, ubound_=ubound)

    elif dist_0 == 'lognorm_':
        dist_kw = lognorm_(**dict_in)
        dist = 'lognorm'
        samples = generate_cfd_q(dist_=dist, dist_kw_=dist_kw, lbound_=lbound, ubound_=ubound)

    elif dist_0 == 'lognorm_mod_':
        dist_kw = lognorm_(**dict_in)
        dist = 'lognorm'
        samples = generate_cfd_q(dist_=dist, dist_kw_=dist_kw, lbound_=lbound, ubound_=ubound)
        samples = 1 - samples

    else:
        raise ValueError('Unknown distribution type {}.'.format(dist))

    samples[samples == np.inf] = ubound
    samples[samples == -np.inf] = lbound

    np.random.shuffle(samples)

    return samples


def dict_unflatten(dict_in: dict):

    dict_out = dict()

    for k in list(dict_in.keys()):
        if ':' in k:
            k1, k2 = k.split(':')

            if k1 in dict_out:
                dict_out[k1][k2] = dict_in[k]
            else:
                dict_out[k1] = dict(k2 = dict_in[k])


def dict_flatten(dict_in: dict):

    dict_out = dict()

    for k in list(dict_in.keys()):
        if isinstance(dict_in[k], dict):
            for kk, vv in dict_in[k].items():
                dict_out[f'{k}:{kk}'] = vv
        else:
            dict_out[k] = dict_in[k]
            
    return dict_in


def main(x: dict, num_samples: int):

    dict_out = dict()

    for k, v in x.items():

        if isinstance(v, float) or isinstance(v, int) or isinstance(v, np.float):
            dict_out[k] = np.full((num_samples,), v, dtype=float)

        elif isinstance(v, str):
            dict_out[k] = np.full((num_samples,), v, dtype=np.dtype('U{:d}'.format(len(v))))
        
        elif isinstance(v, np.ndarray) or isinstance(v, list):
            dict_out[k] = list(np.full((num_samples, len(v)), v, dtype=float))

        elif isinstance(v, dict):
            if 'dist' in v:
                try:
                    dict_out[k] = random_variable_generator(v, num_samples)
                except KeyError:
                    raise('Missing parameters in input variable {}.'.format(k))
            else:
                raise ValueError('Unknown input data type for {}.'.format(k))
        else:
            raise TypeError('Unknown input data type for {}.'.format(k))

    dict_out['index'] = np.arange(0, num_samples, 1)

    df_out = pd.DataFrame.from_dict(dict_out, orient='columns')

    return df_out


def _test_random_variable_generator():

    dict_in = dict(
        v1=np.pi,
        v2='hello world.',
        v3=np.array([0, 1, 2]),
        v4=dict(
            dist='uniform_',
            ubound=10,
            lbound=-1
        ),
        v5=dict(
            dist='norm_',
            ubound=5+1,
            lbound=5-1,
            mean=5,
            sd=1
        ),
        v6=dict(
            dist='gumbel_r_',
            ubound=2500,
            lbound=50,
            mean=420,
            sd=126
        ),
        v7=dict(
            dist='lognorm_',
            ubound=1,
            lbound=0,
            mean=0.5,
            sd=1,
        ),
        v8=dict(
            dist='lognorm_mod_',
            ubound=1,
            lbound=0,
            mean=0.5,
            sd=1,
        )
    )

    df_out = main(dict_in, 10000)

    print(df_out)


if __name__ == '__main__':
    _test_random_variable_generator()



import numpy as np


def fire(
        time,
        temperature_initial
):

    # INPUTS CHECK
    time = np.array(time, dtype=float)
    time[time < 0] = np.nan

    # SI UNITS -> EQUATION UNITS
    temperature_initial -= 273.15  # [K] -> [C]
    time /= 60.  # [s] - [min]

    # CALCULATE TEMPERATURE BASED ON GIVEN TIME
    temperature = 345. * np.log10(time * 8. + 1.) + temperature_initial
    temperature[temperature == np.nan] = temperature_initial

    # EQUATION UNITS -> SI UNITS
    time *= 60.  # [min] -> [s]
    temperature += 273.15  # [C] -> [K]

    return temperature


if __name__ == '__main__':
    fire_time = np.arange(0, 2*60*60+1, 1)
    fire_temperature = fire(fire_time, 273.15+20)

    print(fire_temperature[fire_time==5*60], fire_temperature[fire_time==10*60], fire_temperature[fire_time==15*60],
          fire_temperature[fire_time == 20 * 60], fire_temperature[fire_time==25*60])




# -*- coding: utf-8 -*-
import matplotlib
from sfeprapy.mc1.mc1_obj import MonteCarlo


matplotlib.use('agg')


def main(path_master_csv:str=None):
    MC = MonteCarlo()
    MC.select_input_file(path_master_csv)
    MC.make_input_param()
    MC.make_mc_params()
    MC.run_mc()
    MC.out_combined_ky()
    MC.out_combined_T()


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# from distutils.core import setup

import os
# To use a consistent encoding
from codecs import open

import setuptools

# Get the long description from the README file
with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'README.md')) as file:
    long_description = file.read()

setuptools.setup(

    name='sfeprapy',

    version='0.6.2',

    description='Structural Fire Engineering - Probabilistic Reliability Assessment',

    author='Yan Fu',

    author_email='fuyans@gmail.com',

    url='https://github.com/fsepy/sfeprapy',

    download_url="https://github.com/fsepy/sfeprapy/archive/master.zip",

    keywords=["fire safety", "structural fire engineering"],

    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
    ],

    long_description='Structural Fire Engineering - Probabilistic Reliability Assessment',

    packages=[
        'sfeprapy', 
        'sfeprapy.dat', 
        'sfeprapy.mc0', 
        'sfeprapy.mcs0', 
        'sfeprapy.mc1', 
        'sfeprapy.dist_fit', 
        'sfeprapy.func',
        'sfeprapy.func.heat_transfer_1d'
        ],

    # entry_points={'console_scripts': ['sfeprapymc = sfeprapy.mc0.__main__:main_args']},

    install_requires=[
        'matplotlib>=2.2.2',
        'numpy>=1.15.0',
        'pandas>=0.23.3',
        'scipy>=1.1.0',
        'seaborn>=0.9.0',
        'tqdm',
    ],

    include_package_data=True,
)





def test():
    import copy

    def test_standard_case():
        from sfeprapy.func.mcs_obj import MCS
        from sfeprapy.mcs0 import EXAMPLE_INPUT_DICT, EXAMPLE_CONFIG_DICT
        from sfeprapy.mcs0.mcs0_calc import teq_main, teq_main_wrapper, mcs_out_post
        from sfeprapy.func.mcs_gen import main as gen
        from scipy.interpolate import interp1d
        import numpy as np
        # increase the number of simulations so it gives sensible results
        mcs_input = EXAMPLE_INPUT_DICT
        for k in list(mcs_input.keys()):
            mcs_input[k]['n_simulations'] = 2500
        # increase the number of threads so it runs faster
        mcs_config = copy.copy(EXAMPLE_CONFIG_DICT)
        mcs_config['n_threads'] = 4
        mcs = MCS()
        mcs.define_problem(data=mcs_input, config=mcs_config)
        mcs.define_stochastic_parameter_generator(gen)
        mcs.define_calculation_routine(teq_main, teq_main_wrapper, mcs_out_post)
        mcs.run_mcs()
        mcs_out = mcs.mcs_out
        teq = mcs_out['solver_time_equivalence_solved'] / 60.
        hist, edges = np.histogram(teq, bins=np.arange(0, 181, 0.5))
        x, y = (edges[:-1] + edges[1:]) / 2, np.cumsum(hist / np.sum(hist))
        print(abs(interp1d(y, x)(0.8) - 60))
        assert abs(interp1d(y, x)(0.8) - 60) < 1

    # test gui version
    def test_gui():
        from sfeprapy.func.mcs_obj import MCS
        from sfeprapy.mcs0.mcs0_calc import teq_main as calc
        from sfeprapy.mcs0.mcs0_calc import teq_main_wrapper as calc_mp
        from sfeprapy.func.mcs_gen import main as gen
        mcs = MCS()
        mcs.define_problem()
        mcs.define_stochastic_parameter_generator(gen)
        mcs.define_calculation_routine(calc, calc_mp)
        mcs.run_mcs()

    # test non-gui version
    def test_arg_dict():
        from sfeprapy.func.mcs_obj import MCS
        from sfeprapy.mcs0 import EXAMPLE_INPUT_DICT, EXAMPLE_CONFIG_DICT
        from sfeprapy.mcs0.mcs0_calc import teq_main as calc
        from sfeprapy.mcs0.mcs0_calc import teq_main_wrapper as calc_mp
        from sfeprapy.func.mcs_gen import main as gen
        for k in list(EXAMPLE_INPUT_DICT.keys()):
            EXAMPLE_INPUT_DICT[k]['n_simulations'] = 2
        mcs = MCS()
        mcs.define_problem(data=EXAMPLE_INPUT_DICT, config=EXAMPLE_CONFIG_DICT)
        mcs.define_stochastic_parameter_generator(gen)
        mcs.define_calculation_routine(calc, calc_mp)
        mcs.run_mcs()

    import time
    test_standard_case()
    time.sleep(0.5)
    test_arg_dict()
    time.sleep(0.5)
    test_gui()
    time.sleep(0.5)


if __name__ == '__main__':
    import warnings
    warnings.filterwarnings('ignore')
    test()
